IMPLEMENTATION MODULE GELookup;

        (********************************************************)
        (*                                                      *)
        (*      Web Family Tree: module to find records in      *)
        (*                  a GEDCOM file                       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            7 July 2001                     *)
        (*  Last edited:        18 July 2001                    *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings;

FROM OurTypes IMPORT
    (* type *)  IDString, DataString, FilenameString, Line;

FROM Indexes IMPORT
    (* type *)  Index,
    (* proc *)  OpenIndex, CloseIndex, GetOneLine, PositionToRecord,
                DeleteIndexFile, Match;

FROM TextBuffers IMPORT
    (* type *)  Buffer,
    (* proc *)  OpenForReading, TBFileOpened, CloseTB,
                TBSetPosition, TBStartPosition;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST Nul = CHR(0);  CtrlZ = CHR(26);

TYPE
    CharSet = SET OF CHAR;

    Database = POINTER TO
                   RECORD
                       TB: Buffer;
                       name: FilenameString;
                       index: Index;
                   END (*RECORD*);

    (* Representation of a record, including subsidiary records. *)

    RecordTree = POINTER TO LineRecord;

    LineRecord = RECORD
                     next, down: RecordTree;
                     this:       Line;
                 END (*RECORD*);

CONST Digits = CharSet {'0'..'9'};

(************************************************************************)
(*                   OPENING AND CLOSING A DATABASE                     *)
(************************************************************************)

PROCEDURE OpenDatabase (VAR (*OUT*) DB: Database;
                            DatabaseName: FilenameString): BOOLEAN;

    (* Opens a GEDCOM file. *)

    VAR filename: FilenameString;

    BEGIN
        NEW (DB);
        DB^.name := DatabaseName;
        Strings.Assign ("data\", filename);
        Strings.Append (DatabaseName, filename);
        Strings.Append (".GED", filename);
        DB^.TB := OpenForReading (filename);
        IF NOT TBFileOpened (DB^.TB) THEN
            CloseTB (DB^.TB);
            DISPOSE (DB);
            RETURN FALSE;
        ELSE
            DB^.index := OpenIndex (DB^.TB, DatabaseName);
            RETURN TRUE;
        END (*IF*);
    END OpenDatabase;

(************************************************************************)

PROCEDURE CloseDatabase (VAR (*INOUT*) DB: Database);

    (* Closes a GEDCOM file. *)

    BEGIN
        CloseTB (DB^.TB);
        CloseIndex (DB^.index);
        DISPOSE (DB);
    END CloseDatabase;

(************************************************************************)

PROCEDURE WriteDatabaseName (DB: Database);

    (* Writes the name of this database. *)

    BEGIN
        WriteString (DB^.name);
    END WriteDatabaseName;

(************************************************************************)

PROCEDURE HeadMatch (VAR (*INOUT*) string: ARRAY OF CHAR;
                              template: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff a leading substring of string matches template, *)
    (* modulo character case.  If there is a match, the matched         *)
    (* substring and any following spaces are removed from string.      *)

    VAR j: CARDINAL;

    BEGIN
        j := 0;
        LOOP
            IF (j > HIGH(template)) OR (template[j] = Nul) THEN
                EXIT (*LOOP*);
            END (*IF*);
            IF (j > HIGH(string)) OR (string[j] = Nul)
                    OR (CAP(string[j]) <> CAP(template[j])) THEN
                RETURN FALSE;
            END (*IF*);
            INC (j);
        END (*LOOP*);
        WHILE string[j] = ' ' DO
            INC (j);
        END (*WHILE*);
        Strings.Delete (string, 0, j);
        RETURN TRUE;
    END HeadMatch;

(************************************************************************)

PROCEDURE CrudeSeek (DB: Database;  ID: IDString);

    (* This does the job of SeekToMatchingID (see below) in the case    *)
    (* where an index file lookup fails and we have to go through the   *)
    (* file from the beginning.                                         *)

    VAR thisline: Line;  level, pos: CARDINAL;
        tag: IDString;  found: BOOLEAN;

    BEGIN
        TBSetPosition (DB^.TB, TBStartPosition(DB^.TB));
        LOOP
            level := GetOneLine (DB^.TB, thisline);
            IF (thisline[0] = Nul) OR (thisline[0] = CtrlZ) THEN EXIT(*LOOP*) END(*IF*);
            IF level = 0 THEN
                IF thisline[0] = '@' THEN
                    Strings.FindNext ('@', thisline, 1, found, pos);
                    IF found THEN
                        Strings.Extract (thisline, 1, pos-1, tag);
                        IF ID[0] = Nul THEN
                            REPEAT
                                INC (pos);
                            UNTIL thisline[pos] <> ' ';
                            Strings.Delete (thisline, 0, pos);
                            IF HeadMatch (thisline, 'INDI') THEN
                                EXIT (*LOOP*);
                            END (*IF*);
                        ELSIF Match (tag, ID) THEN
                            EXIT (*LOOP*);
                        END (*IF*);
                    END (*IF*);
                END (*IF*);
            END (*IF*);
        END (*LOOP*);
    END CrudeSeek;

(************************************************************************)

PROCEDURE SeekToMatchingID (DB: Database;  ID: IDString);

    (* Assumption: the file is already open.  We position the file to   *)
    (* the start of the line just after the level 0 line that           *)
    (* matches the ID.                                                  *)

    (* Special case: If ID is the null string then we match the         *)
    (* first INDI record in the file.                                   *)

    VAR thisline: Line;  level: CARDINAL;  found: BOOLEAN;

    BEGIN
        found := PositionToRecord (DB^.index, ID);
        IF found THEN
            level := GetOneLine (DB^.TB, thisline);
            found := level = 0;
        END (*IF*);
        IF NOT found THEN
            DeleteIndexFile (DB^.index);
            CrudeSeek (DB, ID);
        END (*IF*);
    END SeekToMatchingID;

(************************************************************************)

PROCEDURE NonEmpty (T: RecordTree): BOOLEAN;

    (* Returns TRUE iff T is not NIL. *)

    BEGIN
        RETURN T <> NIL;
    END NonEmpty;

(************************************************************************)

PROCEDURE LoadRecordN (TB: Buffer;  VAR (*OUT*) RecTree: RecordTree;
                       N: CARDINAL;  VAR (*INOUT*) NextLevel: CARDINAL;
                       VAR (*INOUT*) NextLine: Line);

    (* On entry, NextLine holds the line just read from the file, and   *)
    (* NextLevel is the level of that line.  We load the level N and    *)
    (* greater lines into RecTree, stripping off the level numbers.     *)
    (* On exit, NextLevel and NextLine are the values for the first     *)
    (* line after the ones we've used.                                  *)

    VAR previous, current: RecordTree;

    BEGIN
        RecTree := NIL;  previous := NIL;

        WHILE (NextLevel <> MAX(CARDINAL)) AND (NextLevel >= N) DO

            NEW (current);
            WITH current^ DO
                next := NIL;
                down := NIL;
            END (*WITH*);

            IF N < NextLevel THEN

                (* Missing record at this level, insert a blank one. *)

                current^.this := "";

            ELSE

                current^.this := NextLine;
                NextLevel := GetOneLine (TB, NextLine);

            END (*IF*);

            IF previous = NIL THEN
                RecTree := current;
            ELSE
                previous^.next := current;
            END (*IF*);

            (* Add the lower-level records, if any. *)

            LoadRecordN (TB, current^.down, N+1, NextLevel, NextLine);

            previous := current;

        END (*WHILE*);

    END LoadRecordN;

(************************************************************************)

PROCEDURE LoadRecord (DB: Database;  VAR (*OUT*) RecTree: RecordTree);

    (* Assumption: the file is already positioned at the start of the   *)
    (* first level 1 line.  We load the level 1 and greater lines into  *)
    (* RecTree, stripping off the level numbers.                        *)

    VAR NextLevel: CARDINAL;  NextLine: Line;

    BEGIN
        NextLevel := GetOneLine (DB^.TB, NextLine);
        LoadRecordN (DB^.TB, RecTree, 1, NextLevel, NextLine);
    END LoadRecord;

(************************************************************************)

PROCEDURE StartReading (DB: Database;  VAR (*OUT*) NextLevel: CARDINAL;
                                          VAR (*OUT*) Lookahead: Line);

    (* Initialisation needed before the first call to LoadNextRecord.   *)

    BEGIN
        TBSetPosition (DB^.TB, TBStartPosition(DB^.TB));
        NextLevel := GetOneLine (DB^.TB, Lookahead);
    END StartReading;

(************************************************************************)

PROCEDURE LoadNextINDIRecord (DB: Database;  VAR (*OUT*) ID: IDString;
                                        VAR (*OUT*) RecTree: RecordTree;
                                        VAR (*INOUT*) NextLevel: CARDINAL;
                                        VAR (*INOUT*) Lookahead: Line);

    (* Loads the next record for an individual into RecTree.  Non-INDI  *)
    (* records are skipped.  On successful return RecTree holds the     *)
    (* level 1 and higher lines of the record.  If there is no next     *)
    (* INDI record, we return with RecTree = NIL.                       *)
    (* Assumption: NextLevel and Lookahead have been "seeded" by a      *)
    (* previous call to LoadNextRecord or StartReading.                 *)

    VAR found: BOOLEAN;  pos: CARDINAL;

    BEGIN
        found := FALSE;
        REPEAT
            IF NextLevel = MAX(CARDINAL) THEN
                RecTree := NIL;
                found := TRUE;
            ELSIF NextLevel = 0 THEN
                IF HeadMatch (Lookahead, 'TRLR') THEN
                    RecTree := NIL;
                    found := TRUE;
                ELSIF Lookahead[0] = '@' THEN
                    Strings.FindNext ('@', Lookahead, 1, found, pos);
                    IF NOT found THEN
                        pos := LENGTH(Lookahead);
                    END (*IF*);
                    Strings.Extract (Lookahead, 1, pos-1, ID);
                    IF pos < LENGTH(Lookahead) THEN
                        REPEAT
                            INC (pos);
                        UNTIL Lookahead[pos] <> ' ';
                    END (*IF*);
                    Strings.Delete (Lookahead, 0, pos);
                    IF HeadMatch (Lookahead, 'INDI') THEN
                        NextLevel := GetOneLine (DB^.TB, Lookahead);
                        LoadRecordN (DB^.TB, RecTree, 1, NextLevel, Lookahead);
                        found := TRUE;
                    ELSE
                        found := FALSE;
                    END (*IF*);
                END (*IF*);
            END (*IF*);
            IF NOT found THEN
                NextLevel := GetOneLine (DB^.TB, Lookahead);
            END (*IF*);
        UNTIL found;
    END LoadNextINDIRecord;

(************************************************************************)

PROCEDURE DiscardTree (VAR (*INOUT*) RecTree: RecordTree);

    (* Releases the storage taken by the tree. *)

    VAR current, next: RecordTree;

    BEGIN
        current := RecTree;
        WHILE current <> NIL DO
            DiscardTree (current^.down);
            next := current^.next;
            DISPOSE (current);
            current := next;
        END (*WHILE*);
        RecTree := NIL;
    END DiscardTree;

(************************************************************************)

PROCEDURE DisplayRawLinesN (T: RecordTree;  N: CARDINAL);

    (* Displays the contents of the tree, labelled with level N and     *)
    (* below.                                                           *)

    VAR j: CARDINAL;

    BEGIN
        WHILE T <> NIL DO
            WriteString ("<br>");
            FOR j := 1 TO 4*N DO
                WriteString ("&nbsp;");
            END (*FOR*);
            WriteString (T^.this);
            WriteLn;
            DisplayRawLinesN (T^.down, N+1);
            T := T^.next;
        END (*WHILE*);
    END DisplayRawLinesN;

(************************************************************************)

PROCEDURE DisplayRawLines (RecTree: RecordTree);

    (* Displays the contents of the tree. *)

    BEGIN
        DisplayRawLinesN (RecTree, 1);
    END DisplayRawLines;

(************************************************************************)

PROCEDURE ExtractSubrecord (keyword: ARRAY OF CHAR;
                            VAR (*INOUT*) T: RecordTree;
                            VAR (*OUT*) data: Line;
                            VAR (*OUT*) subtree: RecordTree): BOOLEAN;

    (* Finds the top-level record, if any, in T that starts with        *)
    (* keyword.  If not found, data is the empty string, subtree=NIL,   *)
    (* T is left unchanged, and the function result is FALSE.           *)
    (* If found, the record and its subsidiary records are removed      *)
    (* from T and the function result is TRUE.  The returned value of   *)
    (* 'data' is the top-level line with keyword and leading spaces     *)
    (* stripped.  The returned value of subtree gives the lower-level   *)
    (* records that were also extracted.                                *)

    VAR previous, current: RecordTree;

    BEGIN
        previous := NIL;  current := T;
        LOOP
            IF current = NIL THEN
                data[0] := Nul;  subtree := NIL;
                RETURN FALSE;
            END (*IF*);
            data := current^.this;
            IF HeadMatch (data, keyword) THEN
                subtree := current^.down;
                IF previous = NIL THEN
                    T := current^.next;
                ELSE
                    previous^.next := current^.next;
                END (*IF*);
                DISPOSE (current);
                RETURN TRUE;
            END (*IF*);
            previous := current;  current := current^.next;
        END (*LOOP*);
    END ExtractSubrecord;

(************************************************************************)

PROCEDURE GetField (keyword: ARRAY OF CHAR;  VAR (*INOUT*) T: RecordTree;
                                  VAR (*OUT*) data: DataString): BOOLEAN;

    (* Like ExtractSubrecord, but discards the sublist. *)

    VAR subtree: RecordTree;

    BEGIN
        IF ExtractSubrecord (keyword, T, data, subtree) THEN
            DiscardTree (subtree);
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END (*IF*);
    END GetField;

(************************************************************************)

PROCEDURE GetName (VAR (*INOUT*) T: RecordTree;
                             VAR (*OUT*) name: DataString): BOOLEAN;

    (* Removes the NAME record from T, returns the name. *)

    VAR j, j1, j2: CARDINAL;  found: BOOLEAN;

    BEGIN

        IF GetField ('NAME', T, name) THEN

            (* Uppercase the part between '/' delimiters. *)

            Strings.FindNext ('/', name, 0, found, j1);
            IF found THEN
                Strings.Delete (name, j1, 1);
                Strings.FindNext ('/', name, j1, found, j2);
                IF found THEN
                    Strings.Delete (name, j2, 1);
                ELSE
                    j2 := LENGTH(name);
                END (*IF*);
                IF j2 > j1 THEN
                    FOR j := j1 TO j2-1 DO
                        name[j] := CAP(name[j]);
                    END (*FOR*);
                END (*IF*);
            END (*IF*);
            RETURN TRUE;

        ELSE
            RETURN FALSE;
        END (*IF*);

    END GetName;

(************************************************************************)

PROCEDURE WritePersonLink2 (DB: Database;  PersonID: IDString;
                                    VAR (*INOUT*) PersonData: RecordTree);

    (* Like WritePersonLink, but for the case where the caller  *)
    (* has already loaded the information for this person into  *)
    (* PersonData.                                              *)

    VAR HisName: DataString;  private: BOOLEAN;

    BEGIN
        private := GetField ('RESN PRIVACY', PersonData, HisName);

            (* Privacy restriction: we can report the name but not      *)
            (* provide a link.                                          *)

        IF NOT private THEN
            WriteString ('<a href="wft.cmd?D=');
            WriteString (DB^.name);
            WriteString (';P=');
            WriteString (PersonID);
            WriteString ('">');
        END (*IF*);

        (* Now fill in the person's name. *)

        IF (NOT GetName (PersonData, HisName)) OR (HisName[0] = Nul) THEN
            HisName := "(name not recorded)";
        END (*IF*);
        WriteString (HisName);

        IF NOT private THEN
            WriteString ('</a>');
        END (*IF*);

    END WritePersonLink2;

(************************************************************************)

PROCEDURE WritePersonLink (DB: Database;  PersonID: IDString);

    (* Turns PersonID into an HTML reference, and writes it to  *)
    (* standard output.                                         *)

    VAR T: RecordTree;

    BEGIN
        SeekToMatchingID (DB, PersonID);
        LoadRecord (DB, T);
        WritePersonLink2 (DB, PersonID, T);
        DiscardTree (T);
    END WritePersonLink;

(************************************************************************)

END GELookup.

