IMPLEMENTATION MODULE Indexes;

        (********************************************************)
        (*                                                      *)
        (*      Web Family Tree: module to deal with the        *)
        (*                    index file                        *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            9 July 2001                     *)
        (*  Last edited:        18 July 2001                    *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings;

FROM OurTypes IMPORT
    (* type *)  FilenameString, Line, IDString;

FROM TextBuffers IMPORT
    (* type *)  Buffer,
    (* proc *)  OpenForReading, TBFileOpened, CloseTB,
                TBSetPosition, TBStartPosition, TBCurrentPosition,
                TBReadRaw, TBReadLine;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId, FilePos,
    (* proc *)  OpenNewFile, CloseFile, DeleteFile,
                WriteRaw, FWriteString;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST Nul = CHR(0);  CtrlZ = CHR(26);

TYPE
    Index = POINTER TO
                RECORD
                    TB: Buffer;
                    dataTB: Buffer;
                    file: FilenameString;
                END (*RECORD*);

(************************************************************************)
(*                       READING FROM INPUT FILE                        *)
(************************************************************************)

PROCEDURE ExtractLevel (VAR (*INOUT*) line: Line): CARDINAL;

    (* Takes one line, trims its level from the front and returns the   *)
    (* level as the function result.                                    *)

    TYPE CharSet = SET OF CHAR;

    CONST Digits = CharSet {'0'..'9'};

    VAR j, level: CARDINAL;

    BEGIN
        level := 0;
        j := 0;
        WHILE line[j] = ' ' DO
            INC (j);
        END (*WHILE*);
        IF line[0] IN Digits THEN
            REPEAT
                level := 10*level + (ORD(line[j]) - ORD('0'));
                INC (j);
            UNTIL NOT (line[j] IN Digits);
        ELSE
            level := MAX(CARDINAL);
        END (*IF*);
        WHILE line[j] = ' ' DO
            INC (j);
        END (*WHILE*);
        IF j > 0 THEN
            Strings.Delete (line, 0, j);
        END (*IF*);
        RETURN level;
    END ExtractLevel;

(************************************************************************)

PROCEDURE GetOneLine (TB: Buffer;  VAR (*OUT*) result: Line): CARDINAL;

    (* Reads one line, trims its level from the front and returns the   *)
    (* level as the function result.                                    *)

    BEGIN
        TBReadLine (TB, result);
        RETURN ExtractLevel(result);
    END GetOneLine;

(************************************************************************)

PROCEDURE GetFirstLine (TB: Buffer;  VAR (*OUT*) Lookahead: Line);

    (* Goes to the beginning of the file, loads Lookahead in            *)
    (* preparation for the first call to GetNextLine.                   *)

    BEGIN
        TBSetPosition (TB, TBStartPosition(TB));
        TBReadLine (TB, Lookahead);
    END GetFirstLine;

(************************************************************************)

PROCEDURE GetNextLine (TB: Buffer;  VAR (*OUT*) result: Line;
                               VAR (*INOUT*) Lookahead: Line): CARDINAL;

    (* Copies Lookahead to result, and then loads the following line    *)
    (* into Lookahead.                                                  *)

    BEGIN
        result := Lookahead;
        TBReadLine (TB, Lookahead);
        RETURN ExtractLevel(result);
    END GetNextLine;

(************************************************************************)
(*                 BUILDING AND REBUILDING AN INDEX FILE                *)
(************************************************************************)

PROCEDURE DeleteIndexFile (IX: Index);

    (* Deletes the index file if it exists.  *)

    BEGIN
        CloseTB (IX^.TB);
        DeleteFile (IX^.file);
    END DeleteIndexFile;

(************************************************************************)

PROCEDURE RebuildIndexFile (IX: Index);

    (* Creates an index file, after deleting the old one if it exists.  *)

    VAR SrcPos: FilePos;
        thisline: Line;  level, pos: CARDINAL;
        tag: IDString;  found: BOOLEAN;
        cid: ChanId;

    BEGIN
        DeleteIndexFile (IX);
        cid := OpenNewFile (IX^.file);
        IF cid <> NoSuchChannel THEN
            TBSetPosition (IX^.dataTB, TBStartPosition(IX^.dataTB));
            LOOP
                SrcPos := TBCurrentPosition (IX^.dataTB);
                level := GetOneLine (IX^.dataTB, thisline);
                IF (thisline[0] = Nul) OR (thisline[0] = CtrlZ) THEN EXIT(*LOOP*) END(*IF*);
                IF level = 0 THEN
                    IF thisline[0] = '@' THEN
                        Strings.FindNext ('@', thisline, 1, found, pos);
                        IF found THEN
                            Strings.Extract (thisline, 1, pos-1, tag);
                            WriteRaw (cid, SrcPos, SIZE(SrcPos));
                            FWriteString (cid, tag);
                            WriteRaw (cid, Nul, 1);
                        END (*IF*);
                    END (*IF*);
                END (*IF*);
            END (*LOOP*);
            CloseFile (cid);
        END (*IF*);
        IX^.TB := OpenForReading (IX^.file);
    END RebuildIndexFile;

(************************************************************************)

PROCEDURE OpenIndex (DataFile: Buffer;  DatabaseName: FilenameString): Index;

    (* Opens an index file. *)

    VAR IX: Index;

    BEGIN
        NEW (IX);
        IX^.dataTB := DataFile;
        Strings.Assign ("data\", IX^.file);
        Strings.Append (DatabaseName, IX^.file);
        Strings.Append (".IDX", IX^.file);
        IX^.TB := OpenForReading (IX^.file);
        IF NOT TBFileOpened (IX^.TB) THEN
            RebuildIndexFile (IX);
        END (*IF*);
        RETURN IX;
    END OpenIndex;

(************************************************************************)

PROCEDURE CloseIndex (VAR (*INOUT*) IX: Index);

    (* Closes an index file. *)

    BEGIN
        IF IX <> NIL THEN
            IF TBFileOpened (IX^.TB) THEN
                CloseTB (IX^.TB);
            END (*IF*);
            DISPOSE (IX);
        END (*IF*);
    END CloseIndex;

(************************************************************************)

PROCEDURE Match (string1, string2: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff string1 and string2 are equal, modulo character *)
    (* case.                                                            *)

    BEGIN
        Strings.Capitalize(string1);
        Strings.Capitalize(string2);
        RETURN Strings.Equal (string1, string2);
    END Match;

(************************************************************************)

PROCEDURE LoadOneIndexRecord (TB: Buffer;  VAR (*OUT*) ThisPos: FilePos;
                                   VAR (*OUT*) ThisID: IDString): BOOLEAN;

    (* Reads an index record from cid. *)

    VAR amount, j: CARDINAL;

    BEGIN
        TBReadRaw (TB, ThisPos, SIZE(FilePos), amount);
        IF amount <> SIZE(FilePos) THEN
            RETURN FALSE;
        END (*IF*);

        j := 0;
        REPEAT
            TBReadRaw (TB, ThisID[j], 1, amount);
            INC (j);
        UNTIL (amount = 0) OR (ThisID[j-1] = Nul) OR (j >= SIZE(IDString));

        RETURN TRUE;

    END LoadOneIndexRecord;

(************************************************************************)

PROCEDURE PositionToRecord (IX: Index;  ID: IDString): BOOLEAN;

    (* Positions the database file to the start of the level 0 record   *)
    (* for the given ID.                                                *)

    VAR found: BOOLEAN;  ThisPos: FilePos;  ThisID: IDString;

    BEGIN
        (* As a proof of concept, for now I'll do the search by a       *)
        (* linear search through the index file.                        *)

        found := TBFileOpened (IX^.TB);
        IF NOT found THEN
            RebuildIndexFile (IX);
            found := TBFileOpened (IX^.TB);
        END (*IF*);
        IF found THEN
            TBSetPosition (IX^.TB, TBStartPosition(IX^.TB));
            LOOP
                found := LoadOneIndexRecord (IX^.TB, ThisPos, ThisID);
                IF NOT found THEN EXIT(*LOOP*) END(*IF*);
                found := Match (ThisID, ID);
                IF found THEN
                    TBSetPosition (IX^.dataTB, ThisPos);
                    EXIT(*LOOP*);
                END(*IF*);
            END (*LOOP*);
        END (*IF*);

        RETURN found;

    END PositionToRecord;

(************************************************************************)

END Indexes.

