IMPLEMENTATION MODULE Person;

        (********************************************************)
        (*                                                      *)
        (*      Web Family Tree: module to look up and          *)
        (*           display the data for a person              *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            6 July 2001                     *)
        (*  Last edited:        16 July 2001                    *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings;

FROM OurTypes IMPORT
    (* type *)  IDString, DataString, FilenameString, Line;

FROM GELookup IMPORT
    (* type *)  Database, RecordTree,
    (* proc *)  OpenDatabase, CloseDatabase,
                SeekToMatchingID, LoadRecord, DiscardTree,
                GetName, GetField, ExtractSubrecord, DisplayRawLines,
                WritePersonLink, WritePersonLink2, WriteDatabaseName,
                StartReading, LoadNextINDIRecord, NonEmpty;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(********************************************************************************)

CONST Nul = CHR(0);

TYPE
    (* Information about one person.                       *)
    (*      ID      the ID in the GEDCOM file              *)
    (*      DB      the database we're looking up          *)
    (*      Tree    result of looking up the GEDCOM file   *)

    PersonData = POINTER TO RECORD
                     ID: IDString;
                     DB: Database;
                     Tree: RecordTree;
                 END (*RECORD*);

(************************************************************************)

PROCEDURE GetPersonData (DatabaseName: FilenameString;  PersonID: IDString;
                                       VAR (*OUT*) result: PersonData): BOOLEAN;

    (* Looks up the database for person with identifier PersonID, fills *)
    (* the result record with the information.                          *)

    BEGIN
        NEW (result);
        result^.ID := PersonID;
        IF NOT OpenDatabase (result^.DB, DatabaseName) THEN
            DISPOSE (result);
            RETURN FALSE;
        END (*IF*);
        SeekToMatchingID (result^.DB, PersonID);
        LoadRecord (result^.DB, result^.Tree);
        RETURN TRUE;
    END GetPersonData;

(**********************************************************************)

PROCEDURE DiscardPersonData (VAR (*INOUT*) data: PersonData);

    (* Disposes of the PersonData information. *)

    BEGIN
        IF data <> NIL THEN
            CloseDatabase (data^.DB);
            DiscardTree (data^.Tree);
            DISPOSE (data);
        END (*IF*);
    END DiscardPersonData;

(***********************************************************************)

PROCEDURE SendLine (line: ARRAY OF CHAR);

    (* Sends a string, terminated with CRLF, to standard output. *)

    BEGIN
        WriteString (line);  WriteLn;
    END SendLine;

(************************************************************************)

PROCEDURE WriteDateAndPlace (VAR (*INOUT*) info: RecordTree);

    (* On entry, info is supposed to contain date and place information.*)
    (* We write this to standard output, and dispose of the tree.       *)

    VAR datum: DataString;

    BEGIN
        IF GetField ('DATE', info, datum) THEN
            WriteString (datum);
        ELSE
            WriteString ("(date not recorded)");
        END (*IF*);
        IF GetField ('PLAC', info, datum) THEN
            WriteString ("<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
            WriteString (datum);
        END (*IF*);
        DiscardTree (info);
    END WriteDateAndPlace;

(************************************************************************)

PROCEDURE ExtractTag (string: ARRAY OF CHAR;  VAR (*OUT*) tag: IDString);

    (* Extracts the substring of string enclosed by @ delimiters. *)

    VAR pos1, pos2: CARDINAL;  found: BOOLEAN;

    BEGIN

        Strings.FindNext ('@', string, 0, found, pos1);
        IF found THEN
            Strings.FindNext ('@', string, pos1+1, found, pos2);
            IF NOT found THEN
                pos2 := LENGTH (string);
            END (*IF*);
            Strings.Extract (string, pos1+1, pos2-pos1-1, tag);
        ELSE
            tag := "";
        END (*IF*);

    END ExtractTag;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Writes a decimal number to standard output. *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE WriteMarriage (M: CARDINAL;  DB: Database;  FamID: IDString);

    (* Writes the details of marriage M. *)

    VAR T, subtree: RecordTree;  PersonID: IDString;
        string: DataString;

    BEGIN
        WriteString ("<h3>Marriage ");
        IF M > 1 THEN
            WriteCard (M);
        END (*IF*);
        WriteString ("</h3>");

        SeekToMatchingID (DB, FamID);
        LoadRecord (DB, T);

        (* Date and place of marriage. *)

        IF ExtractSubrecord ('MARR', T, string, subtree) THEN
            IF NonEmpty(subtree) THEN
                WriteString ("<P>Date: ");
                WriteDateAndPlace (subtree);
                WriteLn;
                DiscardTree (subtree);
            END (*IF*);
        END (*IF*);

        (* Husband *)

        IF GetField ('HUSB', T, string) THEN
            ExtractTag (string, PersonID);
            WriteString ("<br>Husband: ");
            WritePersonLink (DB, PersonID);
            WriteLn;
        END (*IF*);

        (* Wife *)

        IF GetField ('WIFE', T, string) THEN
            ExtractTag (string, PersonID);
            WriteString ("<br>Wife: &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
            WritePersonLink (DB, PersonID);
            WriteLn;
        END (*IF*);

        (* Children *)

        IF GetField ('CHIL', T, string) THEN
            WriteString ("<p>Children:");
            WriteLn;
            REPEAT
                WriteString ("<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
                ExtractTag (string, PersonID);
                WritePersonLink (DB, PersonID);
                WriteLn;
            UNTIL NOT GetField ('CHIL', T, string);
        END (*IF*);

        (* Miscellaneous details. *)

        IF NonEmpty(T) THEN
            WriteString ("<h4>Other marriage details</h4>");
            DisplayRawLines (T);
        END (*IF*);

        DiscardTree (T);

    END WriteMarriage;

(************************************************************************)

PROCEDURE WriteParents (DB: Database;  FamID: IDString);

    (* Writes the names of the two partners to the given family ID,     *)
    (* labelled as father and mother.                                   *)

    VAR T: RecordTree;  FatherID, MotherID: IDString;
        string: DataString;

    BEGIN
        SeekToMatchingID (DB, FamID);
        LoadRecord (DB, T);
        IF GetField ('HUSB', T, string) THEN
            ExtractTag (string, FatherID);
            WriteString ("<p>Father: &nbsp;");
            WritePersonLink (DB, FatherID);
        END (*IF*);
        IF GetField ('WIFE', T, string) THEN
            ExtractTag (string, MotherID);
            WriteString ("<br>Mother: ");
            WritePersonLink (DB, MotherID);
        END (*IF*);
        DiscardTree (T);
    END WriteParents;

(************************************************************************)

PROCEDURE DisplayPerson (data: PersonData);

    (* Displays the information in the PersonData record. *)

    (********************************************************************)

    PROCEDURE MakeButton (option: CHAR;  label: ARRAY OF CHAR);

        (* Puts a button on the web page. *)

        BEGIN
            WriteString ('<form method="post" action="wft.cmd?D=');
            WriteDatabaseName (data^.DB);
            WriteString (';P=');
            WriteString (data^.ID);
            WriteString (';V=');
            WriteChar (option);
            SendLine ('">');
            SendLine ('<input type=hidden name="dummy" value="">');
            WriteString ('<input type="submit" value="');
            WriteString (label);
            WriteString ('">');
            SendLine ('</form>');
        END MakeButton;

    (********************************************************************)

    VAR datum: DataString;  subtree: RecordTree;
        ID: IDString;  M: CARDINAL;

    BEGIN
        SendLine ('<body text="#000000" bgcolor="#FFFFCC" link="#0000EE" vlink="#551A8B" alink="#FF0000">');

        (* The option buttons. *)

        WriteString ('<TABLE BORDER=0 CELLPADDING=10 CELLSPACING=0 RULES="ROWS" >');
        WriteString ('<TR ALIGN=CENTER VALIGN=TOP>');

        WriteString ('<TD>');
        MakeButton ('D', 'Descendants');
        WriteString ('</TD>');

        WriteString ('<TD>');
        MakeButton ('A', 'Ancestors');
        WriteString ('</TD>');

        WriteString ('<TD>');
        MakeButton ('E', 'Everyone');
        WriteString ('</TD>');

        SendLine ('</TR></TABLE>');

        (* Use name as heading. *)

        WriteString ("<h2>");
        IF NOT GetName (data^.Tree, datum) THEN
            datum := "No name";
        END (*IF*);
        WriteString (datum);
        SendLine ("</h2>");

        (* Sex. *)

        IF GetField ('SEX', data^.Tree, datum) THEN
            WriteString ("Sex: ");
            WriteString (datum);
            WriteLn;
        END (*IF*);

        (* Date and place of birth. *)

        IF ExtractSubrecord ('BIRT', data^.Tree, datum, subtree) THEN
            IF NonEmpty(subtree) THEN
                WriteString ("<P>Born: ");
                WriteDateAndPlace (subtree);
                WriteLn;
                DiscardTree (subtree);
            END (*IF*);
        END (*IF*);

        (* Date and place of death. *)

        IF ExtractSubrecord ('DEAT', data^.Tree, datum, subtree) THEN
            IF NonEmpty(subtree) THEN
                WriteString ("<P>Died: ");
                WriteDateAndPlace (subtree);
                WriteLn;
                DiscardTree (subtree);
            END (*IF*);
        END (*IF*);

        (* Parents. *)

        IF GetField ('FAMC', data^.Tree, datum) THEN
            ExtractTag (datum, ID);
            WriteParents (data^.DB, ID);
            WriteLn;
        END (*IF*);

        (* Marriages. *)

        M := 1;
        WHILE GetField ('FAMS', data^.Tree, datum) DO
            ExtractTag (datum, ID);
            WriteMarriage (M, data^.DB, ID);
            INC (M);
        END (*WHILE*);

        (* Display what's left in raw form. *)

        IF NonEmpty(data^.Tree) THEN
            WriteString ("<h3>Other details</h3>");
            DisplayRawLines (data^.Tree);
            DiscardTree (data^.Tree);
        END (*IF*);

    END DisplayPerson;

(************************************************************************)

PROCEDURE AppendCard (N: CARDINAL;  VAR (*INOUT*) str: ARRAY OF CHAR);

    (* Converts N to decimal, appends it to str. *)

    VAR j: CARDINAL;

    BEGIN
        IF N > 9 THEN
            AppendCard (N DIV 10, str);
            N := N MOD 10;
        END (*IF*);
        j := LENGTH(str);
        str[j] := CHR(ORD('0') + N);
        str[j+1] := Nul;
    END AppendCard;

(************************************************************************)

PROCEDURE RDisplayDescendants (prefix: DataString;  level: CARDINAL;
                                DB: Database;
                                ID: IDString;
                                  VAR (*INOUT*) Info: RecordTree);

    (* Displays the names of this person and all direct descendants. *)

    VAR datum, string, newprefix: DataString;  M, N, j: CARDINAL;
        PersonID, FamilyID: IDString;
        TFamily, TChild: RecordTree;

    BEGIN

        WriteString ("<br>");
        FOR j := 1 TO level DO
            WriteString ("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
        END (*FOR*);
        WriteString (prefix);
        WriteString ("&nbsp;&nbsp;&nbsp;");
        WritePersonLink2 (DB, ID, Info);
        Strings.Append ('.', prefix);
        N := 1;

        (* For each marriage ... *)

        M := 1;
        WHILE GetField ('FAMS', Info, datum) DO
            ExtractTag (datum, FamilyID);

            SeekToMatchingID (DB, FamilyID);
            LoadRecord (DB, TFamily);

            (* Extract and display the children of this marriage. *)

            WHILE GetField ('CHIL', TFamily, string) DO
                newprefix := prefix;
                AppendCard (N, newprefix);
                ExtractTag (string, PersonID);
                SeekToMatchingID (DB, PersonID);
                LoadRecord (DB, TChild);
                RDisplayDescendants (newprefix, level+1, DB, PersonID, TChild);
                WriteLn;
                DiscardTree (TChild);
                INC (N);
            END (*WHILE*);

            DiscardTree (TFamily);
            INC (M);

        END (*WHILE*);

    END RDisplayDescendants;

(************************************************************************)

PROCEDURE DisplayDescendants (data: PersonData);

    (* Displays the names of this person and all direct descendants. *)

    BEGIN
        SendLine ("<h2>Direct descendants</h2>");
        RDisplayDescendants ("1", 1, data^.DB, data^.ID, data^.Tree);
        DiscardTree (data^.Tree);
    END DisplayDescendants;

(************************************************************************)

PROCEDURE RDisplayAncestors (prefix: DataString;  level: CARDINAL;
                                DB: Database;
                                ID: IDString;
                                  VAR (*INOUT*) Info: RecordTree);

    (* Displays the names of this person and all direct ancestors. *)

    VAR datum, string, newprefix: DataString;  j: CARDINAL;
        PersonID, FamilyID: IDString;
        TFamily, TChild: RecordTree;

    BEGIN

        WriteString ("<br>");
        FOR j := 1 TO level DO
            WriteString ("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
        END (*FOR*);
        WriteString (prefix);
        WriteString ("&nbsp;&nbsp;&nbsp;");
        WriteLn;
        WritePersonLink2 (DB, ID, Info);

        (* Look up the family, then get husband and wife. *)

        IF GetField ('FAMC', Info, datum) THEN
            ExtractTag (datum, FamilyID);

            SeekToMatchingID (DB, FamilyID);
            LoadRecord (DB, TFamily);

            (* Extract and display the husband and wife of this marriage. *)

            IF GetField ('HUSB', TFamily, string) THEN
                newprefix := prefix;
                Strings.Append ('M', newprefix);
                ExtractTag (string, PersonID);
                SeekToMatchingID (DB, PersonID);
                LoadRecord (DB, TChild);
                RDisplayAncestors (newprefix, level+1, DB, PersonID, TChild);
                WriteLn;
                DiscardTree (TChild);
            END (*IF*);

            IF GetField ('WIFE', TFamily, string) THEN
                newprefix := prefix;
                Strings.Append ('F', newprefix);
                ExtractTag (string, PersonID);
                SeekToMatchingID (DB, PersonID);
                LoadRecord (DB, TChild);
                RDisplayAncestors (newprefix, level+1, DB, PersonID, TChild);
                WriteLn;
                DiscardTree (TChild);
            END (*IF*);

            DiscardTree (TFamily);

        END (*IF*);

    END RDisplayAncestors;

(************************************************************************)

PROCEDURE DisplayAncestors (data: PersonData);

    (* Displays the names of this person and all direct ancestors. *)

    VAR sex: DataString;

    BEGIN
        SendLine ("<h2>Ancestors in the direct line</h2>");
        IF NOT GetField ('SEX', data^.Tree, sex) THEN
           sex := '?';
        END (*IF*);
        RDisplayAncestors (sex, 1, data^.DB, data^.ID, data^.Tree);
        DiscardTree (data^.Tree);
    END DisplayAncestors;

(************************************************************************)

PROCEDURE DisplayEveryone (dummy: PersonData);

    (* Displays the names of everyone in the database.  The person      *)
    (* mentioned in the argument is unimportant, but we need some       *)
    (* person record so that we can identify the database.              *)

    VAR NextLevel, count: CARDINAL;  Lookahead: Line;
        T: RecordTree;  DB: Database;  success: BOOLEAN;
        ID: IDString;

    BEGIN
        DB := dummy^.DB;
        SendLine ("<h2>All individuals in the database</h2>");
        StartReading (DB, NextLevel, Lookahead);
        count := 0;
        REPEAT
            LoadNextINDIRecord (DB, ID, T, NextLevel, Lookahead);
            success := NonEmpty(T);
            IF success THEN
                INC (count);
                WriteString ("<br>");
                WritePersonLink2 (DB, ID, T);
                WriteLn;
                DiscardTree (T);
            END (*IF*);
        UNTIL NOT success;
        WriteString ("<p>Total: ");  WriteCard (count);
        WriteLn;
    END DisplayEveryone;

(************************************************************************)
(*                           INITIALISATION                             *)
(************************************************************************)

END Person.

